--[[
-- Copyright (c) 2012  Anton Savoshchanka (aka AntonioModer)
--
-- This software is provided 'as-is', without any express or implied
-- warranty.  In no event will the authors be held liable for any damages
-- arising from the use of this software.
--
-- Permission is granted to anyone to use this software for any purpose,
-- including commercial applications, and to alter it and redistribute it
-- freely, subject to the following restrictions:
--
-- 1. The origin of this software must not be misrepresented; you must not
--    claim that you wrote the original software. If you use this software
--    in a product, an acknowledgment in the product documentation would be
--    appreciated but is not required.
-- 2. Altered source versions must be plainly marked as such, and must not be
--    misrepresented as being the original software.
-- 3. This notice may not be removed or altered from any source distribution.
--]]

-- P.S.: comments in Russian

-- ###################################################################

-- tableFind ------------------------------------------------------------------------------
--       
-- tab -  
-- str -   
function tableFind(tab, str)
	local str = tostring(str);
	local tab = tab;					-- speed run
	local newTab = {};
	
	for k, v in pairs(tab) do			-- find
		if tostring(v) == str then
			newTab[#newTab+1] = k;		-- insert
		end
	end
	return newTab;
end

-- tableFindAndDelete ------------------------------------------------------------------------------
-- tab -     .  -    !
function tableFindAndDelete(tab, str)
	local tabFind = tableFind(tab, str);
	-- print(tabFind[3])
	for i =  #tabFind, 1, -1 do				--    
		table.remove(tab, tabFind[i]);
	end
	tabFind = nil;
end

--  rayCastGetBody()
-- tableFindAndDeleteDuplicate ------------------------------------------------------------------------------
-- tab -     .  -    !
function tableFindAndDeleteDuplicate(tab)
	--      --
	local tabFind = {};
	for k, v in pairs(tab) do
		tabFind = tableFind(tab, v);
		for i = #tabFind, 2, -1 do				--    2 (..   ,  )
			table.remove(tab, tabFind[i]);
		end
	end	
	tabFind = nil;
end

--  sliceConvex
--      :
-- : tab[1] = objIndex; ...
function rayCastGetBody(rayHitList1)
	local tabObj = {};
	local rayPNBody = 0;
	local rayHitList1 = rayHitList1;
	-- print("---> "..#rayHitList1)

	for k, v in pairs(rayHitList1) do 				--    
		rayPNBody = (v.fixture:getUserData()).objNameIndex;
		table.insert(tabObj, rayPNBody);
	end
	
	--      --
	tableFindAndDeleteDuplicate(tabObj)
	
	return tabObj;
end

--  sliceConvex
-- return table,    ,    ray
function rayCastPointsInObj(rayHitList1, rayHitList2, fobjNameIndex)
	local fobjNameIndex = tostring(fobjNameIndex);
	local rayHitList1 = rayHitList1;
	local rayHitList2 = rayHitList2;
	local rayPNBody = 0;	
	local tabPointsInObj = {};
	for k, v in pairs(rayHitList1) do
		rayPNBody = (v.fixture:getUserData()).objNameIndex;
		
		if rayPNBody == fobjNameIndex then															--      objNameIndex
			tabPointsInObj[#tabPointsInObj+1] = v.x;												-- insert
			tabPointsInObj[#tabPointsInObj+1] = v.y;
		end
	end	
	for k, v in pairs(rayHitList2) do
		rayPNBody = (v.fixture:getUserData()).objNameIndex;
		if rayPNBody == fobjNameIndex then															--      objNameIndex
			tabPointsInObj[#tabPointsInObj+1] = v.x;												-- insert
			tabPointsInObj[#tabPointsInObj+1] = v.y;
		end
	end	
	-- print(unpack(tabPointsInObj));																	-- DEBUG
	return tabPointsInObj;
end

--  sliceConvex
--   
-- return
-- TODO:
-- test:
function endPointsOfVector(tabPoints)
	local tabPoints = tabPoints;
	local pEnd1 = {};							--    (    "")
	pEnd1.x, pEnd1.y = 0, 0;
	local pEnd2 = {};							--    (    "")
	pEnd2.x, pEnd2.y = 0, 0;
	local tabMax = {};							--   x,y  
	tabMax.x, tabMax.y = 0, 0;
	local tabMin = {};							--   x,y  
	tabMin.x, tabMin.y = 0, 0;
	local tabX = {};
	local tabY = {};
	local tabVectors = {}; 						--  
	
	
	for i=1, #tabPoints/2 do 					--   
		tabVectors[i] = {};						--  
		tabVectors[i].x = tabPoints[i+(i-1)]	-- x
		tabVectors[i].y = tabPoints[i+i]		-- y
	end	
	
	for i=1, #tabPoints, 2 do 					--   X
		tabX[#tabX+1] = tabPoints[i];
	end
	for i=2, #tabPoints, 2 do 					--   Y
		tabY[#tabY+1] = tabPoints[i];
	end	
	tabMax.x = math.max(unpack(tabX));
	tabMax.y = math.max(unpack(tabY));
	-- tabMin.x = math.min(unpack(tabX));
	tabMin.y = math.min(unpack(tabY));
	
	pEnd1.x = tabMax.x;							--     
	for i=1, #tabVectors do						--   
		if tabVectors[i].x < pEnd1.x  then		--  
			pEnd1.x = tabVectors[i].x;
			pEnd1.y = tabVectors[i].y;			-- 
		end	
			
		if tabVectors[i].x > pEnd2.x  then		--  
			pEnd2.x = tabVectors[i].x;
			pEnd2.y = tabVectors[i].y;			-- 
		end
	end 
	if pEnd1.x == pEnd2.x then					--   
		pEnd1.y = tabMin.y;						--  -  
		pEnd2.y = tabMax.y;						--  -  
	end
	
	-- DEBUG --
	-- print(pEnd1.x, pEnd1.y)
	-- print(pEnd2.x, pEnd2.y)
	
	return pEnd1.x, pEnd1.y, pEnd2.x, pEnd2.y;
end

function findSameVectors(x1,y1, x2,y2, x3,y3, x4,y4)
	local x1,y1, x2,y2, x3,y3, x4,y4 = 0,0,0,0,0,0,0,0;
	if math.angle(x1,y1, x2,y2) == math.angle(x3,y3, x4,y4) then
		return x1,y1, x2,y2;
	elseif math.angle(x2,y2, x1,y1) == math.angle(x3,y3, x4,y4) then
		return x2,y2, x1,y1;
	end	
end


-- ###################################################################

-- return numb xCenter, yCenter
function findCenterDraw(obj)
	local obj = obj;
	local cNewX, cNewY = 0, 0;
	local vX, vY = 0, 0;
	local tabShapeVert = convexGetVertex(obj);	
	
	for v=1, #tabShapeVert/2 do
		vX, vY = obj.physBody:getWorldPoint(tabShapeVert[v+(v-1)], tabShapeVert[v+v])
		cNewX = cNewX + vX; -- sum global
		cNewY = cNewY + vY;

		if v == #tabShapeVert/2 then
			cNewX = cNewX/v;
			cNewY = cNewY/v;
		end
	end
	-- print('='..cNewX)
	return cNewX, cNewY;	
end

-- return numb xCenter, yCenter
function findCenter(tabShapeVert)
	local tabShapeVert = tabShapeVert;
	local cNewX, cNewY = 0, 0;
	local vX, vY = 0, 0;
	
	for v=1, #tabShapeVert/2 do
		vX, vY = tabShapeVert[v+(v-1)], tabShapeVert[v+v]; 			-- local
		cNewX = cNewX + vX; -- sum
		cNewY = cNewY + vY;
		-- print(cNewX)
		if v == #tabShapeVert/2 then
			cNewX = cNewX/v;
			cNewY = cNewY/v;
		end
	end
	-- print('='..cNewX)
	return cNewX, cNewY;
end

function convexGetVertex(obj)
	local obj = obj;
	--    --
	local tabVerConvex = {};
	local x1, y1 = obj.physShapePoly[1]:getPoints();
	table.insert(tabVerConvex, x1);																--  
	table.insert(tabVerConvex, y1);	
	
	for tr=1, #obj.physShapePoly do														--     obj
		local x1, y1, x2, y2 = obj.physShapePoly[tr]:getPoints();
		table.insert(tabVerConvex, x2);															--  
		table.insert(tabVerConvex, y2);
	end
	
	local x1, y1, x2, y2, x3, y3 = obj.physShapePoly[#obj.physShapePoly]:getPoints(); --   
	table.insert(tabVerConvex, x3);																-- 3     (  )
	table.insert(tabVerConvex, y3);
	
	return tabVerConvex;
end

-- v. 0.8.31
-- [FIX] Slice and delete only polygonal shape.
-- [ADD] Point of the mass of the object at the center of object.
-- [ADD] Draw the direction of motion of the object.
-- [ADD] Create new center for new body.
-- [ADD] The transfer of "speed" of the old object to new objects.
function sliceConvex(ray1Start, ray1End, Ray1HitList, Ray2HitList)
	print("start sliceConvex function")																-- DEBUG
	
	local tabCollBody = rayCastGetBody(Ray1HitList);												--    obj  ray
	local pEnd1x, pEnd1y, pEnd2x, pEnd2y = 0,0,0,0; 												--   obj
	local tabPointsL = {};																			--  
	local tabPointsR = {};																			-- ... 
	-- print(tabCollBody)	
	
	-- print("NO BUG")
	for obj=1, #tabCollBody do																		--    ray

		-- Body Left --
		-- tabCollBody[obj] = "objNameIndex" --
		local objIndex = tabEntitys[tabCollBody[obj]];
		
		if objIndex.physShapeType == 3 or objIndex.physShapeType == 4 then							-- slice only polygons
		
			local tabCollPointBody = rayCastPointsInObj(Ray1HitList, Ray2HitList, tabCollBody[obj]);	--      obj
			
			pEnd1x, pEnd1y, pEnd2x, pEnd2y = endPointsOfVector(tabCollPointBody);							--      obj
			
			--  R1  R2 --
			local R1, R2 = {}, {};
			R1.x, R1.y, R2.x, R2.y = 0, 0, 0, 0;
			if math.floor(math.deg(math.angle(pEnd1x, pEnd1y, pEnd2x, pEnd2y))) == math.floor(math.deg(math.angle(ray1.x1,ray1.y1, ray1.x2,ray1.y2))) then
				R1.x, R1.y = pEnd1x, pEnd1y;
				R2.x, R2.y = pEnd2x, pEnd2y;
			elseif math.floor(math.deg(math.angle(pEnd2x, pEnd2y, pEnd1x, pEnd1y))) == math.floor(math.deg(math.angle(ray1.x1,ray1.y1, ray1.x2,ray1.y2))) then
				R1.x, R1.y = pEnd2x, pEnd2y;
				R2.x, R2.y = pEnd1x, pEnd1y;
			end	
			
			--    --
			local tabVerConvex = {};																	-- table with index
			local x1, y1 = objIndex.physShapePoly[1]:getPoints();
			table.insert(tabVerConvex, x1);																--  
			table.insert(tabVerConvex, y1);	

			for tr=1, #objIndex.physShapePoly do														--     obj
				local x1, y1, x2, y2 = objIndex.physShapePoly[tr]:getPoints();
				table.insert(tabVerConvex, x2);															--  
				table.insert(tabVerConvex, y2);
			end
			
			local x1, y1, x2, y2, x3, y3 = objIndex.physShapePoly[#objIndex.physShapePoly]:getPoints(); --   
			table.insert(tabVerConvex, x3);																-- 3     (  )
			table.insert(tabVerConvex, y3);
			
			--    --
			if #tabPointsL ~= 0 then
				for i=#tabPointsL, 1, -1  do															-- clear
					table.remove(tabPointsL, i);
				end		
			end	
			for vCon = 1, #tabVerConvex/2 do	--    
				-- print("start for tabVerConvex")														-- DEBUG
				local vWorldX, vWorldY = objIndex.physBody:getWorldPoint(tabVerConvex[vCon+(vCon-1)], tabVerConvex[vCon+vCon]);	--     
				
				if math.isLeftOrRightOrCenter(ray1Start.x, ray1Start.y, ray1End.x, ray1End.y, vWorldX, vWorldY) < 0 then		-- if point left
					-- print(#tabPointsL)
					table.insert(tabPointsL, tabVerConvex[vCon+(vCon-1)])				-- x local
					table.insert(tabPointsL, tabVerConvex[vCon+vCon])					-- y local	

				end
			end			

			-- new algoritm, for convex --
			local R1LocalX, R1LocalY = objIndex.physBody:getLocalPoint(R1.x, R1.y);							--  r1  r2   
			local R2LocalX, R2LocalY = objIndex.physBody:getLocalPoint(R2.x, R2.y);
			local tableObjL = {};
			
			-- print(objIndex.physBody:getWorldPoints(unpack(tabPointsL)))
			if #tabPointsL > 2 then
				local borderR = false;																	--       
				for vCon=1, #tabPointsL/2 do
					if vCon == #tabPointsL/2 then
						-- print(borderR)
						if borderR == false then
							table.insert(tableObjL, tabPointsL[vCon+(vCon-1)])					-- x N local
							table.insert(tableObjL, tabPointsL[vCon+vCon])						-- y N local
							
							table.insert(tableObjL, R2LocalX)							-- R2x local
							table.insert(tableObjL, R2LocalY)							-- R2y local

							table.insert(tableObjL, R1LocalX)							-- R1, point TrN1,   ...  1 local
							table.insert(tableObjL, R1LocalY)							-- R1, point TrN1, ...		 local							
						else
							table.insert(tableObjL, tabPointsL[vCon+(vCon-1)])					-- x N local
							table.insert(tableObjL, tabPointsL[vCon+vCon])						-- y N local	
						end
					else

						local pNLWX, pNLWY = objIndex.physBody:getWorldPoint(tabPointsL[vCon+(vCon-1)], tabPointsL[vCon+vCon]);				-- point N left world
						local pN1LWX, pN1LWY = objIndex.physBody:getWorldPoint(tabPointsL[(vCon+(vCon-1))+2], tabPointsL[(vCon+vCon)+2]);	-- point N+1 left world	-- BUG
						-- print(pNLWX, pNLWY, pN1LWX, pN1LWY)
						-- print((vCon+(vCon-1))+1)
						
						-- print(math.isLeftOrRightOrCenter(pNLWX, pNLWY, pN1LWX, pN1LWY, R2.x, R2.y) >= 0)
						if math.isLeftOrRightOrCenter(pNLWX, pNLWY, pN1LWX, pN1LWY, R2.x, R2.y) >= 0 then		--   (  )
							table.insert(tableObjL, tabPointsL[vCon+(vCon-1)])					-- x N local
							table.insert(tableObjL, tabPointsL[vCon+vCon])						-- y N local						
						else
							borderR = true;
							table.insert(tableObjL, tabPointsL[vCon+(vCon-1)])					-- x N local
							table.insert(tableObjL, tabPointsL[vCon+vCon])						-- y N local
							
							table.insert(tableObjL, R2LocalX)							-- R2x local
							table.insert(tableObjL, R2LocalY)							-- R2y local

							table.insert(tableObjL, R1LocalX)							-- R1, point TrN1,   ...  1 local
							table.insert(tableObjL, R1LocalY)							-- R1, point TrN1, ...		 local					
						end
					end
				end
			else
				-- R2+R1+ 
				
				table.insert(tableObjL, R2LocalX)							-- R2x local
				table.insert(tableObjL, R2LocalY)							-- R2y local

				table.insert(tableObjL, R1LocalX)							-- R1, point TrN1,   ...  1 local
				table.insert(tableObjL, R1LocalY)							-- R1, point TrN1, ...		 local
				
				table.insert(tableObjL, tabPointsL[1])									-- 1-  
				table.insert(tableObjL, tabPointsL[2])		
			end
			
			
			-- new body
			
			local trNBodyX, trNBodyY  = objIndex.physBody:getLocalPoint(objIndex.physBody:getX(), objIndex.physBody:getY()); -- local

			-- add new center for new obj
			local cNewX, cNewY = findCenter(tableObjL); -- local
			local diffX = cNewX - trNBodyX;			 -- difference
			local diffY = cNewY - trNBodyY;			 -- difference
		
			
			--      --
			local vX, vY = 0, 0;
			for vCon=1, #tableObjL/2 do
				vX = tableObjL[vCon+(vCon-1)] - diffX;
				vY = tableObjL[vCon+vCon] - diffY;
				tableObjL[vCon+(vCon-1)] = vX;															--          
				tableObjL[vCon+vCon] = vY;
			end				
			
			
			local trX, trY  = objIndex.physBody:getWorldPoint(cNewX, cNewY);
			testBody = tostring("_"..globalEntityCount+1);
			tabEntitys[testBody] = Entity:new(trX, trY, testBody);     
			tabEntitys[testBody]:setPhysBody("dynamic", 4, 0, 0, tableObjL, tabEntitys[testBody]);	
			tabEntitys[testBody].physBody:setAngle(objIndex.physBody:getAngle());
			-- tabEntitys[testBody].physBody:setAngle(math.angle(trNBodyX,trNBodyY, cNewX, cNewY) );
			tabEntitys[testBody].physBody:setAwake(false);
			
			-- point of mass = in center of body
			local massX, massY, mass, inertia = tabEntitys[testBody].physBody:getMassData();
			tabEntitys[testBody].physBody:setMassData( 0, 0, mass, inertia)

			--   --
			local vx, xy = objIndex.physBody:getLinearVelocityFromLocalPoint( 0, 0 )
			tabEntitys[testBody].physBody:applyLinearImpulse(vx/5, xy/5)	
			-- print(vx, xy)
			
			-- Body Right --
			
			--    --
			if #tabPointsR ~= 0 then
				for i=#tabPointsR, 1, -1  do															-- clear
					table.remove(tabPointsR, i);
				end		
			end	
			for vCon = 1, #tabVerConvex/2 do	--    
				-- print("start for tabVerConvex")														-- DEBUG
				local vWorldX, vWorldY = objIndex.physBody:getWorldPoint(tabVerConvex[vCon+(vCon-1)], tabVerConvex[vCon+vCon]);	--     
				
				if math.isLeftOrRightOrCenter(ray1Start.x, ray1Start.y, ray1End.x, ray1End.y, vWorldX, vWorldY) > 0 then		-- if point right
					-- print(#tabPointsR)
					table.insert(tabPointsR, tabVerConvex[vCon+(vCon-1)])				-- x local
					table.insert(tabPointsR, tabVerConvex[vCon+vCon])					-- y local	

				end
			end			
			
			-- new algoritm, for convex --
			local tableObjR = {};
			
			-- print(objIndex.physBody:getWorldPoints(unpack(tabPointsR)))
			if #tabPointsR > 2 then
				local borderR = false;																	--       
				for vCon=1, #tabPointsR/2 do
					if vCon == #tabPointsR/2 then
						-- print(borderR)
						if borderR == false then
							table.insert(tableObjR, tabPointsR[vCon+(vCon-1)])					-- x N local
							table.insert(tableObjR, tabPointsR[vCon+vCon])						-- y N local

							table.insert(tableObjR, R1LocalX)							-- R1, point TrN1,   ...  1 local
							table.insert(tableObjR, R1LocalY)							-- R1, point TrN1, ...		 local	
							
							table.insert(tableObjR, R2LocalX)							-- R2x local
							table.insert(tableObjR, R2LocalY)							-- R2y local
							
						else
							table.insert(tableObjR, tabPointsR[vCon+(vCon-1)])					-- x N local
							table.insert(tableObjR, tabPointsR[vCon+vCon])						-- y N local	
						end
					else
						
						local pNLWX, pNLWY = objIndex.physBody:getWorldPoint(tabPointsR[vCon+(vCon-1)], tabPointsR[vCon+vCon]);				-- point N left world
						local pN1LWX, pN1LWY = objIndex.physBody:getWorldPoint(tabPointsR[(vCon+(vCon-1))+2], tabPointsR[(vCon+vCon)+2]);	-- point N+1 left world	-- BUG
						-- print(pNLWX, pNLWY, pN1LWX, pN1LWY)
						
						if math.isLeftOrRightOrCenter(pNLWX, pNLWY, pN1LWX, pN1LWY, R2.x, R2.y) >= 0 then		--   (  )
							table.insert(tableObjR, tabPointsR[vCon+(vCon-1)])					-- x N local
							table.insert(tableObjR, tabPointsR[vCon+vCon])						-- y N local						
						else
							borderR = true;
							table.insert(tableObjR, tabPointsR[vCon+(vCon-1)])					-- x N local
							table.insert(tableObjR, tabPointsR[vCon+vCon])						-- y N local
							
							table.insert(tableObjR, R1LocalX)							-- R1, point TrN1,   ...  1 local
							table.insert(tableObjR, R1LocalY)							-- R1, point TrN1, ...		 local	
							
							table.insert(tableObjR, R2LocalX)							-- R2x local
							table.insert(tableObjR, R2LocalY)							-- R2y local
						end
					end
				end
			else
				-- R1+R2+ 
				table.insert(tableObjR, R1LocalX)							-- R1, point TrN1,   ...  1 local
				table.insert(tableObjR, R1LocalY)							-- R1, point TrN1, ...		 local	
				
				table.insert(tableObjR, R2LocalX)							-- R2x local
				table.insert(tableObjR, R2LocalY)							-- R2y local
				
				table.insert(tableObjR, tabPointsR[1])									-- 1-  
				table.insert(tableObjR, tabPointsR[2])		
			end
			
			-- print(#tableObjR)
			-- new body
			
			local trNBodyX, trNBodyY  = objIndex.physBody:getLocalPoint(objIndex.physBody:getX(), objIndex.physBody:getY()); -- local

			-- add new center for new obj
			local cNewX, cNewY = findCenter(tableObjR); -- local
			local diffX = cNewX - trNBodyX;			 -- difference
			local diffY = cNewY - trNBodyY;			 -- difference
		
			
			--      --
			local vX, vY = 0, 0;
			for vCon=1, #tableObjR/2 do
				vX = tableObjR[vCon+(vCon-1)] - diffX;
				vY = tableObjR[vCon+vCon] - diffY;
				tableObjR[vCon+(vCon-1)] = vX;															--          
				tableObjR[vCon+vCon] = vY;
			end				
			
			
			local trX, trY  = objIndex.physBody:getWorldPoint(cNewX, cNewY);
			testBody = tostring("_"..globalEntityCount+1);
			tabEntitys[testBody] = Entity:new(trX, trY, testBody);     
			tabEntitys[testBody]:setPhysBody("dynamic", 4, 0, 0, tableObjR, tabEntitys[testBody]);	
			tabEntitys[testBody].physBody:setAngle(objIndex.physBody:getAngle());
			-- tabEntitys[testBody].physBody:setAngle(math.angle(trNBodyX,trNBodyY, cNewX, cNewY) );
			tabEntitys[testBody].physBody:setAwake(false);
			
			-- point of mass = in center of body
			local massX, massY, mass, inertia = tabEntitys[testBody].physBody:getMassData();
			tabEntitys[testBody].physBody:setMassData( 0, 0, mass, inertia)
			
			--   --
			local vx, xy = objIndex.physBody:getLinearVelocityFromLocalPoint( 0, 0 )
			tabEntitys[testBody].physBody:applyLinearImpulse(vx/5, xy/5)	
			-- print(vx, xy)
		end
	end
	
	for obj = 1, #tabCollBody do
		local objIndex = tabEntitys[tabCollBody[obj]];
		if objIndex.physShapeType == 3 or objIndex.physShapeType == 4 then							-- delete only polygons
			objIndex.physBody:destroy();											    			-- destroy body    !!!
			tabEntitys[objIndex.objNameIndex]	= nil												-- destroy Object  !!!	
		end
	end
	

	
	print("end sliceConvex function")																-- DEBUG
end
















